open GccTypes

let print_comment = ref false

let print_ins c lp ins =
  match ins with
  | LDC n -> Printf.fprintf c "  LDC %d" n
  | LD (a,b) -> Printf.fprintf c "  LD %d %d" a b
  | ADD -> Printf.fprintf c "  ADD"
  | SUB -> Printf.fprintf c "  SUB"
  | MUL -> Printf.fprintf c "  MUL"
  | DIV -> Printf.fprintf c "  DIV"
  | CEQ -> Printf.fprintf c "  CEQ"
  | CGT -> Printf.fprintf c "  CGT"
  | CGTE -> Printf.fprintf c "  CGTE"
  | CONS -> Printf.fprintf c "  CONS"
  | CAR -> Printf.fprintf c "  CAR"
  | CDR -> Printf.fprintf c "  CDR"
  | ATOM -> Printf.fprintf c "  ATOM"
  | SEL (a,b) ->
     let s1, l1 = lp a in
     let s2, l2 = lp b in
     Printf.fprintf c "  SEL %s %s\t; %s1 %s2" l1 l2 s1 s2
  | JOIN -> Printf.fprintf c "  JOIN"
  | LDF n ->
     let s1, l1 = lp n in
     Printf.fprintf c "  LDF %s\t; %s" l1 s1
  | AP n -> Printf.fprintf c "  AP %d" n
  | RTN -> Printf.fprintf c "  RTN"
  | DUM n -> Printf.fprintf c "  DUM %d" n
  | RAP n -> Printf.fprintf c "  RAP %d" n
  | STOP -> Printf.fprintf c "  STOP"
  | TSEL (a,b) ->
     let s1, l1 = lp a in
     let s2, l2 = lp b in
     Printf.fprintf c "  TSEL %s %s\t; %s %s " l1 l2 s1 s2
  | TAP n -> Printf.fprintf c "  TAP %d" n
  | TRAP n -> Printf.fprintf c "  TRAP %d" n
  | ST (a,b) -> Printf.fprintf c "  ST %d %d" a b
  | DBUG -> Printf.fprintf c "  DBUG"
  | BRK -> Printf.fprintf c "  BRK"
  | COMMENT n when !print_comment -> Printf.fprintf c "; %s" n
  | COMMENT n -> ()
  | LABEL n -> Printf.fprintf c "%s:" (snd (lp n))

let print_int_program c p =
  Array.iter
    (fun ins ->
      print_ins c (fun i -> "", string_of_int i) ins;
      Printf.fprintf c "\n"
    ) p

let print_arg_program c p =
  let map = ref StringMap.empty in
  let rec iter n n' =
    if n < Array.length p then
      match p.(n) with
      | LABEL (L s) ->
        map := StringMap.add s n' !map;
        iter (n+1) n'
      | LABEL (I _) -> iter (n+1) n'
      | COMMENT _ -> iter (n+1) n'
      | _ -> iter (n+1) (n' +1)
  in
  iter 0 0;

  let label_printer = function
    | I i -> string_of_int i, ""
    | L s ->
       try
         let n = StringMap.find s !map in
         (Printf.sprintf "%S -> %d" s n)
         ,string_of_int n
       with Not_found ->
         Printf.eprintf
           "Error: could not find definition of label %S\n%!" s;
         "0","error!"
  in

  let rec print_insts labels = function
    | [] -> ()
    | LABEL l :: t ->
       let s = match l with
         | L s -> Printf.sprintf "\t; %d: %s" (StringMap.find s !map) s
         | I i -> Printf.sprintf "\t; %d" i
       in
       print_insts (s :: labels) t
    | COMMENT s :: t ->
       print_insts (("\t; " ^ s) :: labels) t
    | i :: t ->
       print_ins c label_printer i;
       List.iter (fun s -> Printf.fprintf c "%s" s) labels;
       Printf.fprintf c "\n";
       print_insts [] t
  in

  print_insts [] (Array.to_list p)


let print_label_arg_program c p =
  let f = function
    | L s -> s
    | I i -> string_of_int i
  in
  Array.iter
    (fun ins ->
      print_ins c (fun l -> "", f l) ins;
      Printf.fprintf c "\n"
    ) p

let make_label_map code =
  let labels = ref StringMap.empty in
  let rec mark i = function
    | LABEL (L s) ->
       labels := StringMap.add s i !labels;
    | LABEL (I n) ->
       assert false
    | COMMENT _ -> ()
    | _ -> ()
  in
  Array.iteri mark code;
  !labels

(* let make_label_map code = *)
(*   let labels = ref StringMap.empty in *)
(*   let rec mark i = function *)
(*     | [] -> () *)
(*     | LABEL (L s) :: t -> *)
(*        labels := StringMap.add s i !labels; *)
(*        mark i t *)
(*     | LABEL (I n) :: t -> *)
(*        assert(i = n); *)
(*        mark i t *)
(*     | COMMENT _ :: t -> mark i t *)
(*     | instr :: t -> mark (i+1) t *)
(*   in *)
(*   mark 0 (Array.to_list code); *)
(*   !labels *)

let assemble code =
  let labels = ref StringMap.empty in
  let find = function
    | I i -> i
    | L s -> StringMap.find s !labels in
  let rec mark i = function
    | [] -> []
    | LABEL (L s) :: t ->
       labels := StringMap.add s i !labels;
       mark i t
    | LABEL (I n) :: t ->
       assert(i = n);
       mark i t
    | COMMENT _ :: t -> mark i t
    | instr :: t -> instr :: (mark (i+1) t)
  in
  let code = mark 0 (Array.to_list code) in
  let aux : arg ins -> int ins = function
    | (LDC _
      | LD _
      | ADD
      | SUB
      | MUL
      | DIV
      | CEQ
      | CGT
      | CGTE
      | CONS
      | CAR
      | CDR
      | ATOM
      | JOIN
      | AP _
      | RTN
      | DUM _
      | RAP _
      | STOP
      | TAP _
      | TRAP _
      | ST _
      | DBUG
      | BRK) as i -> i
    | SEL (a,b) -> SEL (find a, find b)
    | LDF n -> LDF (find n)
    | TSEL (a,b) ->
       TSEL (find a,find b)
    | COMMENT n -> assert false
    | LABEL n -> assert false
  in
  Array.map aux (Array.of_list code)

let show_ins (ins:arg ins) =
  let lbl = function
    | I i -> string_of_int i
    | L s -> s
  in
  match ins with
  | LDC n -> Printf.sprintf "  LDC %d" n
  | LD (a,b) -> Printf.sprintf "  LD %d %d" a b
  | ADD -> Printf.sprintf "  ADD"
  | SUB -> Printf.sprintf "  SUB"
  | MUL -> Printf.sprintf "  MUL"
  | DIV -> Printf.sprintf "  DIV"
  | CEQ -> Printf.sprintf "  CEQ"
  | CGT -> Printf.sprintf "  CGT"
  | CGTE -> Printf.sprintf "  CGTE"
  | CONS -> Printf.sprintf "  CONS"
  | CAR -> Printf.sprintf "  CAR"
  | CDR -> Printf.sprintf "  CDR"
  | ATOM -> Printf.sprintf "  ATOM"
  | SEL (a,b) ->
     Printf.sprintf "  SEL %s %s" (lbl a) (lbl b)
  | JOIN -> Printf.sprintf "  JOIN"
  | LDF n ->
     Printf.sprintf "  LDF %s" (lbl n)
  | AP n -> Printf.sprintf "  AP %d" n
  | RTN -> Printf.sprintf "  RTN"
  | DUM n -> Printf.sprintf "  DUM %d" n
  | RAP n -> Printf.sprintf "  RAP %d" n
  | STOP -> Printf.sprintf "  STOP"
  | TSEL (a,b) ->
     Printf.sprintf "  TSEL %s %s" (lbl a) (lbl b)
  | TAP n -> Printf.sprintf "  TAP %d" n
  | TRAP n -> Printf.sprintf "  TRAP %d" n
  | ST (a,b) -> Printf.sprintf "  ST %d %d" a b
  | DBUG -> Printf.sprintf "  DBUG"
  | BRK -> Printf.sprintf "  BRK"
  | COMMENT n -> Printf.sprintf "  COMMENT %s" n
  | LABEL n -> Printf.sprintf "  LABEL %s" (lbl n)
