external halt : unit -> 'a = "%raise"

(* Composition operators *)

external ( |> ) : 'a -> ('a -> 'b) -> 'b = "%revapply"
external ( @@ ) : ('a -> 'b) -> 'a -> 'b = "%apply"

(* Comparisons *)

external ( = ) : 'a -> 'a -> bool = "%equal"
external ( <> ) : 'a -> 'a -> bool = "%notequal"
external ( < ) : 'a -> 'a -> bool = "%lessthan"
external ( > ) : 'a -> 'a -> bool = "%greaterthan"
external ( <= ) : 'a -> 'a -> bool = "%lessequal"
external ( >= ) : 'a -> 'a -> bool = "%greaterequal"
external compare : 'a -> 'a -> int = "%compare"

let min x y = if x <= y then x else y
let max x y = if x >= y then x else y

external ( == ) : 'a -> 'a -> bool = "%eq"
external ( != ) : 'a -> 'a -> bool = "%noteq"

(* Boolean operations *)

external not : bool -> bool = "%boolnot"
external ( & ) : bool -> bool -> bool = "%sequand"
external ( && ) : bool -> bool -> bool = "%sequand"
external ( or ) : bool -> bool -> bool = "%sequor"
external ( || ) : bool -> bool -> bool = "%sequor"

(* Integer operations *)

external ( ~- ) : int -> int = "%negint"
external ( ~+ ) : int -> int = "%identity"
external succ : int -> int = "%succint"
external pred : int -> int = "%predint"
external ( + ) : int -> int -> int = "%addint"
external ( - ) : int -> int -> int = "%subint"
external ( *  ) : int -> int -> int = "%mulint"
external ( / ) : int -> int -> int = "%divint"
external ( mod ) : int -> int -> int = "%modint"

let abs x = if x >= 0 then x else -x

external ( land ) : int -> int -> int = "%andint"
external ( lor ) : int -> int -> int = "%orint"
external ( lxor ) : int -> int -> int = "%xorint"

let lnot x = x lxor (-1)

(* Character operations -- more in module Char *)

external int_of_char : char -> int = "%identity"
external unsafe_char_of_int : int -> char = "%identity"
(* let char_of_int n = *)
  (* if n < 0 || n > 255 then halt () else unsafe_char_of_int n *)

(* Unit operations *)

external ignore : 'a -> unit = "%ignore"

(* Pair operations *)

external fst : 'a * 'b -> 'a = "%field0"
external snd : 'a * 'b -> 'b = "%field1"

(* List operations -- more in module List *)

let rec ( @ ) l1 l2 =
  match l1 with
    [] -> l2
  | hd :: tl -> hd :: (tl @ l2)

external compare : 'a -> 'a -> int = "caml_compare"

(* References *)

type 'a ref = { mutable contents : 'a }
external ref : 'a -> 'a ref = "%makemutable"
external ( ! ) : 'a ref -> 'a = "%field0"
external ( := ) : 'a ref -> 'a -> unit = "%setfield0"
external incr : int ref -> unit = "%incr"
external decr : int ref -> unit = "%decr"

(* lists *)
let rec list_length_aux len = function
  | [] -> len
  | a::l -> list_length_aux (len + 1) l

let list_length l = list_length_aux 0 l

let list_nth l n =
  if n < 0 then None else
  let rec nth_aux l n =
    match l with
    | [] -> None
    | a::l -> if n = 0 then Some a else nth_aux l (n-1)
  in nth_aux l n

let rec list_rev_append l1 l2 =
  match l1 with
  | [] -> l2
  | a :: l -> list_rev_append l (a :: l2)

let list_rev l = list_rev_append l []

let list_append a b = list_rev_append (list_rev a) b

let (@) = list_append

let rec list_flatten = function
    [] -> []
  | l::r -> l @ list_flatten r

let list_concat = list_flatten

let rec list_map f l =
  match l with
  | [] -> []
  | hd :: tl -> f hd :: list_map f tl

let list_mapi f l =
  let rec mapi l i =
    match l with
    | [] -> []
    | hd :: tl -> f hd i :: mapi tl (i + 1)
  in mapi l 0

let rec list_iter f l =
  match l with
  | [] -> ()
  | hd :: tl -> f hd ; list_iter f tl

let list_iteri f l =
  let rec iteri l i =
    match l with
    | [] -> ()
    | hd :: tl -> f hd i ; iteri tl (i + 1)
  in iteri l 0

let rec list_fold_left f acc l =
  match l with
  | [] -> acc
  | hd :: tl -> list_fold_left f (f acc hd) tl

let rec list_for_all pred l =
  match l with
  | [] -> true
  | hd :: tl -> pred hd && list_for_all pred tl

let rec list_exists pred l =
  match l with
  | [] -> false
  | hd :: tl -> pred hd || list_exists pred tl

let rec list_find pred l =
  match l with
  | [] -> None
  | hd :: tl -> if pred hd then Some hd else list_find pred tl

(*
let rec list_merge cmp l1 l2 =
  match l1, l2 with
  | [], l2 -> l2
  | l1, [] -> l1
  | h1 :: t1, h2 :: t2 ->
      if cmp h1 h2 <= 0
      then h1 :: list_merge cmp t1 l2
      else h2 :: list_merge cmp l1 t2

let rec list_chop k l =
  if k = 0 then l else begin
    match l with
    | x::t -> list_chop (k-1) t
    | _ -> assert false
  end

let list_stable_sort cmp l =
  let rec list_rev_merge l1 l2 accu =
    match l1, l2 with
    | [], l2 -> list_rev_append l2 accu
    | l1, [] -> list_rev_append l1 accu
    | h1::t1, h2::t2 ->
        if cmp h1 h2 <= 0
        then list_rev_merge t1 l2 (h1::accu)
        else list_rev_merge l1 t2 (h2::accu)
  in
  let rec list_rev_merge_list_rev l1 l2 accu =
    match l1, l2 with
    | [], l2 -> list_rev_append l2 accu
    | l1, [] -> list_rev_append l1 accu
    | h1::t1, h2::t2 ->
        if cmp h1 h2 > 0
        then list_rev_merge_list_rev t1 l2 (h1::accu)
        else list_rev_merge_list_rev l1 t2 (h2::accu)
  in
  let rec list_sort n l =
    match n, l with
    | 2, x1 :: x2 :: _ ->
       if cmp x1 x2 <= 0 then [x1; x2] else [x2; x1]
    | 3, x1 :: x2 :: x3 :: _ ->
       if cmp x1 x2 <= 0 then begin
         if cmp x2 x3 <= 0 then [x1; x2; x3]
         else if cmp x1 x3 <= 0 then [x1; x3; x2]
         else [x3; x1; x2]
       end else begin
         if cmp x1 x3 <= 0 then [x2; x1; x3]
         else if cmp x2 x3 <= 0 then [x2; x3; x1]
         else [x3; x2; x1]
       end
    | n, l ->
       let n1 = n asr 1 in
       let n2 = n - n1 in
       let l2 = list_chop n1 l in
       let s1 = list_rev_sort n1 l in
       let s2 = list_rev_sort n2 l2 in
       list_rev_merge_list_rev s1 s2 []
  and list_rev_sort n l =
    match n, l with
    | 2, x1 :: x2 :: _ ->
       if cmp x1 x2 > 0 then [x1; x2] else [x2; x1]
    | 3, x1 :: x2 :: x3 :: _ ->
       if cmp x1 x2 > 0 then begin
         if cmp x2 x3 > 0 then [x1; x2; x3]
         else if cmp x1 x3 > 0 then [x1; x3; x2]
         else [x3; x1; x2]
       end else begin
         if cmp x1 x3 > 0 then [x2; x1; x3]
         else if cmp x2 x3 > 0 then [x2; x3; x1]
         else [x3; x2; x1]
       end
    | n, l ->
       let n1 = n asr 1 in
       let n2 = n - n1 in
       let l2 = list_chop n1 l in
       let s1 = list_sort n1 l in
       let s2 = list_sort n2 l2 in
       list_rev_merge s1 s2 []
  in
  let len = list_length l in
  if len < 2 then l else list_sort len l

let list_sort = list_stable_sort
*)

let list_insert i l =
  let rec rew = function
    | [] -> [i]
    | hd :: tl as l -> if hd < i then hd :: rew tl else i :: l
  in rew l

(* native pairs *)
type 'a pair
external car : ('a * 'b) pair -> 'a pair = "percent_car"
external cdr : ('a * 'b) pair -> 'b pair = "percent_cdr"
external cons : 'a pair -> 'b pair -> ('a * 'b) pair = "percent_cons"

(* OCaml pairs *)
let fst (a, _) = a
let snd (_, b) = b

type value
external value : 'a -> value = "%identity"
external magic : value -> 'a = "%identity"
external car : value -> value = "percent_car"
external cdr : value -> value = "percent_cdr"
external cons : value -> value -> value = "percent_cons"
external atom : value -> bool = "percent_atom"
external debug : 'a -> unit = "percent_debug"

let cadr v = (car (cdr v))
let cdar v = (cdr (car v))
let caar v = (car (car v))
let cddr v = (cdr (cdr v))

let caadr v = (car (car (cdr v)))
let cadar v = (car (cdr (car v)))
let caaar v = (car (car (car v)))
let caddr v = (car (cdr (cdr v)))
let cdadr v = (cdr (car (cdr v)))
let cddar v = (cdr (cdr (car v)))
let cdaar v = (cdr (car (car v)))
let cdddr v = (cdr (cdr (cdr v)))

let caaadr v = (car (car (car (cdr v))))
let caadar v = (car (car (cdr (car v))))
let caaaar v = (car (car (car (car v))))
let caaddr v = (car (car (cdr (cdr v))))
let cadadr v = (car (cdr (car (cdr v))))
let caddar v = (car (cdr (cdr (car v))))
let cadaar v = (car (cdr (car (car v))))
let cadddr v = (car (cdr (cdr (cdr v))))
let cdaadr v = (cdr (car (car (cdr v))))
let cdadar v = (cdr (car (cdr (car v))))
let cdaaar v = (cdr (car (car (car v))))
let cdaddr v = (cdr (car (cdr (cdr v))))
let cddadr v = (cdr (cdr (car (cdr v))))
let cdddar v = (cdr (cdr (cdr (car v))))
let cddaar v = (cdr (cdr (car (car v))))
let cddddr v = (cdr (cdr (cdr (cdr v))))

(* Array *)

module Array = struct
  external get: 'a array -> int -> 'a = "%array_unsafe_get"
  external set: 'a array -> int -> 'a -> unit = "%array_unsafe_set"
end

external array_length: 'a array -> int = "%array_length"
external array_get : 'a array -> int -> 'a = "%array_unsafe_get"
external array_set : 'a array -> int -> 'a -> unit = "%array_unsafe_set"
external array_create: int -> 'a -> 'a array = "caml_make_vect"

let array_init l f =
  let res = array_create l (f 0) in
  for i = 1 to pred l do
    array_set res i (f i)
  done;
  res

let aget = array_get
let aset = array_set

let array_get2 map x y = array_get (array_get map y) x
let array_set2 map x y v = array_set (array_get map y) x v

let buffered_array default =
  let buf = ref (array_init 255 (fun _ -> array_create 255 default)) in
  let buf_prev = ref (array_init 255 (fun _ -> array_create 255 default)) in
  let array_get x y = array_get2 !buf x y in
  let array_get' x y = array_get2 !buf_prev x y in
  let array_set x y = array_set2 !buf x y in
  let swap () =
    let tmp = !buf_prev in
    buf_prev := !buf ;
    buf := tmp
  in
  array_get, array_get', array_set, swap

let array w h default =
  let buf = array_init h (fun _ -> array_create w default) in
  let array_get x y = array_get2 buf x y in
  let array_set x y = array_set2 buf x y in
  array_get, array_set

let marker w h =
  let marks = array_init h (fun _ -> array_create w false) in
  let marked x y = array_get2 marks x y in
  let mark x y = array_set2 marks x y true in
  marked, mark

let debug_bools get w h =
  let map = ref (value 0) in
  for j = h - 1 downto 0 do
    let row = ref (value 0) in
    for i = w - 1 downto 0 do
      row := cons (value (get i j)) !row
    done ;
    map := cons !row !map
  done ;
  debug (cons (value 1664_1664) !map)

let debug_array get w h =
  let map = ref (value 0) in
  for j = h - 1 downto 0 do
    let row = ref (value 0) in
    for i = w - 1 downto 0 do
      row := cons (value (get i j)) !row
    done ;
    map := cons !row !map
  done ;
  debug (cons (value 1665_1665) !map)

