type instr =
  | Jmp of rvalue
  | Set of lvalue * rvalue
  | Arith of arith_op * lvalue * rvalue
  | Cmp of cmp_op * int * rvalue * rvalue
  | Interrupt of int
  | Halt
  | Undefined
 and rvalue =
   | Reg of reg
   | Cst of int
   | Mem_at_reg of reg
   | Mem_at_cst of int
   | Pc
 and lvalue =
   | Set_reg of reg
   | Set_mem_at_reg of reg
   | Set_mem_at_cst of int
 and arith_op =
   | Add | Sub | Mul | Div
   | And | Or | Xor
 and cmp_op =
   | Lt | Eq  | Gt
 and reg = int
 and interrupt = int

let rvalue_of_val (v:value) : rvalue =
  let x = magic (cdr v) in
  match magic (car v) with
  | 0 -> Reg x
  | 1 -> Mem_at_reg x
  | 2 -> Cst x
  | _ -> Mem_at_cst x

let lvalue_of_val (v:value) : lvalue =
  let x = magic (cdr v) in
  match magic (car v) with
  | 0 -> Set_reg x
  | 1 -> Set_mem_at_reg x
  | _ -> Set_mem_at_cst x
  
let extract_constant (v:value) : int = magic (cdr v)

let instr_of_val (v : value) : instr =
  let opcode = magic (car v) in
  let args = cdr v in
  match opcode with
  | 0 -> Jmp (rvalue_of_val (car args))
  | 1 -> Arith (Add, lvalue_of_val (car args), Cst 1)
  | 2 -> Arith (Sub, lvalue_of_val (car args), Cst 1)
  | 3 -> Arith (Add, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 4 -> Arith (Sub, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 5 -> Arith (Mul, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 6 -> match Arith (Div, lvalue_of_val (car args), rvalue_of_val (cadr args)) with Arith (_, _, Cst 0) -> Halt | i -> i
  | 7 -> Arith (And, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 8 -> Arith (Or, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 9 -> Arith (Xor, lvalue_of_val (car args), rvalue_of_val (cadr args))
  | 10 -> Cmp ( Lt, magic (car args),
                rvalue_of_val (cadr args),
                rvalue_of_val (caddr args))
  | 11 -> Cmp ( Eq, magic (car args),
                rvalue_of_val (cadr args),
                rvalue_of_val (caddr args))
  | 12 -> Cmp ( Gt, magic (car args),
                rvalue_of_val (cadr args),
                rvalue_of_val (caddr args))
  | 13 -> Interrupt (magic (car args))
  | _ -> Halt

let gprogram_of_val (l:value) : instr array =
  let a = array_create Halt in
  let rec aux i l =
    if i = 256 || atom l
    then ()
    else (
      aset a i (instr_of_val (car l));
      aux (i+1) (cdr l) )
  in
  aux 0 l; a

let rec ghosts_program_retrieve_rec first (l:value) : instr array list =
  if atom l
  then []
  else let next = gprogram_of_val (car l) in
    if compare (value next) (value first) = 0
    then []
    else next :: (ghosts_program_retrieve_rec first (cdr l))

let ghosts_program_retrieve l =
  if atom l
  then []
  else
  let first = (gprogram_of_val (car l)) in
  first :: (ghosts_program_retrieve_rec first (cdr l))
