open Ghc

type var = int
type table = int

type label = string

type expr =
  | Const of int
  | Op of arith_op * expr * expr
  | Var of int
  | Field of table * expr
  | Ghost_index
  | Place of expr * expr

type largue =
  | Var of int
  | Field of table * expr

type info_inter =
  | Ghost_start_pos
  | Ghost_current_pos
  | Ghost_vit_dir

type statement =
  | Assign of largue * expr
  | Set_dir of expr
  | Lambda_man_pos of largue * largue * int
  | Ghost_info of info_inter * largue * largue * expr
  | Debug
  | Debugv of expr
  | Int2

type statements = statement list

type branch =
  | Cond of (cmp_op * label * expr * expr)
  | Cond2 of (cmp_op * label * label * expr * expr)
  | Goto of label
  | End
  | No_branch
  | Goto_manual of var
  | Label_expr of var * label

type branch' =
  | BCond of (cmp_op * label * rvalue * rvalue)
  | BCond2 of (cmp_op * label * label * rvalue * rvalue)
  | BGoto of label
  | BEnd
  | BNo_branch
  | BGoto_manual of var
  | BLabel_expr of var * label

type block =
  label * statements * branch

type code = block list

let first_reg = 3
(* 0 and 1 are used for interruptions, 2 is reserved for arrays *)

let last_reg = 30

let too_many_colons = "expression trop compliquée... désolé"

let array_reg = 2

let reg i = i + first_reg

let var_pos i =
  let pos = i - 8 in
  if pos > last_reg
  then failwith too_many_colons
  else pos

let rvar i =
  let i = reg i in
  if i < 8
  then Reg i
  else
    let pos = var_pos i in
    Mem_at_cst pos

let lvar i =
  let i = reg i in
  if i < 8
  then Set_reg i
  else
    let pos = var_pos i in
    Set_mem_at_cst pos

let var i =
  last_reg + 1 + i

let rec depth = function
  | Const _ -> 0
  | Op (_, e1, e2) -> 1 + max (depth e1) (depth e2)
  | Var _ -> 0
  | Field (_, e) -> depth e
  | Ghost_index -> 0
  | Place (e1, e2) -> 1 + max (depth e1) (depth e2)

let to_rval = function
  | Set_reg r -> Reg r
  | Set_mem_at_reg r -> Mem_at_reg r
  | Set_mem_at_cst c -> Mem_at_cst c

let equal_lval dst lval = match dst, lval with
  | Reg r1, Set_reg r2 -> r1 = r2
  | Mem_at_reg r1, Set_mem_at_reg r2 -> r1 = r2
  | Mem_at_cst c1, Set_mem_at_cst c2 -> c1 = c2
  | _ -> false

let set_to_rval dst = function
  | Set(x, rval) :: t when equal_lval dst x -> t, rval
  | l -> l, dst

let lval_to_string = function
  | Set_reg i -> Printf.sprintf "reg %i" i
  | Set_mem_at_cst v -> Printf.sprintf "[%i]" v
  | _ -> "other"

let mk_free_reg () =
  let r = ref (-1) in
  fun () ->
  incr r;
  let _ = var_pos !r in
  !r

let gen_expr free_reg dest expr =
  let rec aux dest expr =
    match expr with
    | Const c ->
       [Set (dest, (Cst c))]

    | Op (op, e1, e2) ->
       let tmp1 = free_reg () in
       let tmp2 = free_reg () in
       let e1 = aux (lvar tmp1) e1 in
       let e2 = aux (lvar tmp2) e2 in
       let e2, result2 = set_to_rval (rvar tmp2) e2 in

       Set(dest, rvar tmp1)
       :: Arith(op, lvar tmp1, result2)
       :: (e2 @ e1)

    | Var v ->
       if equal_lval (Mem_at_cst (v)) dest
       then []
       else [Set(dest, Mem_at_cst (v))]

    | Field (table, e) ->
       let e = aux (Set_reg array_reg) e in
       Set(dest, Mem_at_reg(array_reg))
       :: Arith(Add, Set_reg array_reg, Cst table)
       :: e

  | Ghost_index ->
      Set(dest, Reg 0) :: [Interrupt 3]

  | Place (e1,e2) ->
       let tmp1 = free_reg () in
       let tmp2 = free_reg () in
       let e1 = aux (lvar tmp1) e1 in
       let e2 = aux (lvar tmp2) e2 in
       let e2, result2 = set_to_rval (rvar tmp2) e2 in

       Set(dest, Reg 0)
       :: Interrupt 7
       :: Set(Set_reg 0, rvar tmp1)
       :: Set(Set_reg 1, rvar tmp2)
       :: (e2 @ e1)

  in
  List.rev (aux dest expr)

(* let rec comp_expr' dest i expr = *)
(*   (\* Printf.printf "comp_expr %s\n%!" (lval_to_string dest); *\) *)
(*   match expr with *)
(*   | Const c -> *)
(*       [Set (dest, (Cst c))] *)
(*   | Op (op, e1, e2) -> *)
(*      (\* let e2 = comp_expr' (lvar i) (i+1) e2 in *\) *)
(*      (\* let e1 = comp_expr' dest (i+1) e1 in *\) *)
(*      (\* let e2, r = set_to_rval (rvar i) e2 in *\) *)
(*      (\* Arith(op, dest, r) :: (e1 @ e2) *\) *)

(*      let e2 = comp_expr' (lvar i) (i+1) e2 in *)
(*      let e1 = comp_expr' (lvar (i+1)) (i+2) e1 in *)
(*      Set(dest, rvar(i+1)) :: Arith(op, lvar (i+1), rvar i) :: (e1 @ e2) *)

(*   | Var v -> *)
(*       if equal_lval (Mem_at_cst (v)) dest *)
(*       then [] *)
(*       else [Set(dest, Mem_at_cst (v))] *)
(*   | Field (table, e) -> *)
(*       let e = comp_expr' (lvar i) (i+1) e in *)
(*       Set(dest, Mem_at_reg(reg i)) :: (Arith(Add, lvar i, Cst table) :: e) *)
(*   | Ghost_index -> *)
(*       Set(dest, Reg 0) :: [Interrupt 3] *)

(* let comp_expr' dest i e = *)
(*   List.rev (comp_expr' dest i e) *)

(* let comp_expr dest e = *)
(*   comp_expr' dest (reg (-1)) e *)

let make_largue free_reg tmp_reg = function
  | Var v ->
     Set_mem_at_cst v, []
  | Field (t,e) ->
     Set_mem_at_reg tmp_reg,
     gen_expr free_reg (Set_reg tmp_reg) (Op (Add, e, Const t))

(* let make_largue x n = function *)
(*   | Var v -> Set_mem_at_cst v, [] *)
(*   | Field (t,e) -> *)
(*       Set_mem_at_reg x, *)
(*       comp_expr' (Set_reg x) n (Op (Add, e, Const t)) *)

let ghost_interrupt = function
  | Ghost_start_pos -> 4
  | Ghost_current_pos -> 5
  | Ghost_vit_dir -> 6

let comp_statement = function
  (* | Assign (Var v, expr) -> *)
  (*    comp_expr' (Set_reg (reg 0)) 0 expr @ [Set(Set_mem_at_cst v, Reg (reg 0))] *)

  | Assign (largue, expr) ->
     let free_reg = mk_free_reg () in
     let dest, tt = make_largue free_reg (free_reg ()) largue in
     tt @ gen_expr free_reg dest expr


  | Set_dir expr ->
     let free_reg = mk_free_reg () in
     gen_expr free_reg (Set_reg 0) expr @ [Interrupt 0]

  | Lambda_man_pos (l1, l2, i) ->
      let v =
        match i with
        | 1 -> 1
        | 2 -> 2
        | _ -> failwith "bad lambda man"
      in
      let free_reg = mk_free_reg () in
      let tmp1 = free_reg () in
      let tmp2 = free_reg () in
      let dest1, tt1 = make_largue free_reg tmp1 l1 in
      let dest2, tt2 = make_largue free_reg tmp2 l2 in
      tt1 @ tt2 @
      Interrupt v :: Set (dest1, Reg 0) :: [Set (dest2, Reg 1)]
  | Ghost_info (info, l1, l2, index) ->
      let inter = ghost_interrupt info in
      let free_reg = mk_free_reg () in
      let tmp1 = free_reg () in
      let tmp2 = free_reg () in
      let dest1, tt1 = make_largue free_reg tmp1 l1 in
      let dest2, tt2 = make_largue free_reg tmp2 l2 in
      let index = gen_expr free_reg (Set_reg 0) index in
      tt1 @ tt2 @ index @
      Interrupt inter :: Set (dest1, Reg 0) :: [Set (dest2, Reg 1)]
  | Debug ->
      [Interrupt 8]
  | Debugv expr ->
     let free_reg = mk_free_reg () in
     gen_expr free_reg (Set_reg 0) expr @ [Interrupt 8]

  | Int2 -> [Interrupt 2]

let comp_statements l =
  List.flatten (List.map comp_statement l)

let rec last_load rval = function
  | Set(_, (Mem_at_cst x as rval)) :: t when x > last_reg ->
      last_load rval t
  | l -> l, rval

let rec ultimate_load rval = function
  | Set(_, rval) :: t -> ultimate_load rval t
  | l -> l, rval

let prepare_branch = function
  | End -> BEnd, []
  | Goto lbl -> BGoto lbl, []
  | Cond (cmp, lbl, e1, e2) ->
     let free_reg = mk_free_reg () in
     let tmp1 = free_reg () in
     let tmp2 = free_reg () in
     let e1 = gen_expr free_reg (lvar tmp1) e1 in
     let e2 = gen_expr free_reg (lvar tmp2) e2 in
     BCond(cmp, lbl, rvar tmp1, rvar tmp2), e1 @ e2

      (* let e1 = comp_expr' (Set_reg (reg 0)) (reg 0) e1 in *)
      (* let e2 = comp_expr' (Set_reg (reg 1)) (reg 1) e2 in *)
      (* let dest1 = (Reg (reg 0)) in *)
      (* let dest2 = (Reg (reg 1)) in *)
      (* let e1, dest1 = last_load dest1 e1 in *)
      (* let e2, dest2 = ultimate_load dest2 e2 in *)
      (* BCond(cmp, lbl, dest1, dest2), e1 @ e2 *)
  | Cond2 (cmp, lbl1, lbl2, e1, e2) ->
     let free_reg = mk_free_reg () in
     let tmp1 = free_reg () in
     let tmp2 = free_reg () in
     let e1 = gen_expr free_reg (lvar tmp1) e1 in
     let e2 = gen_expr free_reg (lvar tmp2) e2 in
     BCond2(cmp, lbl1, lbl2, rvar tmp1, rvar tmp2), e1 @ e2

      (* let e1 = comp_expr' (Set_reg (reg 0)) (reg 0) e1 in *)
      (* let e2 = comp_expr' (Set_reg (reg 1)) (reg 1) e2 in *)
      (* let dest1 = (Reg (reg 0)) in *)
      (* let dest2 = (Reg (reg 1)) in *)
      (* let e1, dest1 = last_load dest1 e1 in *)
      (* let e2, dest2 = ultimate_load dest2 e2 in *)
      (* BCond2(cmp, lbl1, lbl2, dest1, dest2), e1 @ e2 *)

  | No_branch -> BNo_branch, []
  | Goto_manual v ->
      BGoto_manual v, []
  | Label_expr (v, lbl) ->
      BLabel_expr (v, lbl), []


let comp_block (label, statements, branch) =
  let code = comp_statements statements in
  let br, br_code = prepare_branch branch in
  label, code @ br_code, br

let branch_size = function
  | BEnd -> 1
  | BGoto _ -> 1
  | BCond _ -> 1
  | BCond2 _ -> 2
  | BNo_branch -> 0
  | BGoto_manual _ -> 1
  | BLabel_expr _ -> 1

let build_blocks (blocks:code) =
  let blocks = List.map comp_block blocks in
  let t = Hashtbl.create (List.length blocks) in
  let pos i (label, code, branch) =
    if Hashtbl.mem t label
    then failwith ("duplicate label " ^ label);
    Hashtbl.add t label i;
    List.length code + i + (branch_size branch)
  in
  let (_:int) = List.fold_left pos 0 blocks in
  let get l =
    try Hashtbl.find t l
    with Not_found -> failwith ("unknown label " ^ l)
  in
  let rebuild (label, code, branch) =
    let expr = match branch with
      | BEnd -> [ Halt ]
      | BGoto l -> [ Jmp (Cst (get l)) ]
      | BCond (cmp, lbl, r1, r2) ->
          [ Cmp(cmp, get lbl, r1, r2) ]
      | BCond2 (cmp, lbl1, lbl2, r1, r2) ->
          [ Cmp(cmp, get lbl1, r1, r2); Jmp (Cst (get lbl2)) ]
      | BNo_branch -> []
      | BGoto_manual v -> [ Jmp (Mem_at_cst v) ]
      | BLabel_expr (v, lbl) -> [ Set (Set_mem_at_cst v, Cst (Hashtbl.find t lbl)) ]
    in
    code @ expr
  in
  List.flatten (List.map rebuild blocks)

let ( *^ ) x y = Op(Mul, x, y)
let ( -^ ) x y = Op(Sub, x, y)
let ( =^ ) a b = Assign  (a, b)
let ( /^ ) x y = Op (Div, x, y)
let ( +^ ) x y = Op (Add, x, y)
let ( &&^ ) x y = Op (And, x, y)

let factorielle : code =
  let n = var 0 in
  let acc = var 1 in
  [
    "entry",
    [Var acc =^ Const 1],
    No_branch;

    "loop",
    [],
    Cond(Eq, "exit", Var n, Const 0);

    "body",
    [Var acc =^ Var n *^ Var acc;
     Var n   =^ Var n -^ Const 1],
    Goto "loop";

    "exit",
    [],
    End
  ]

(* let fact = build_blocks factorielle *)


let mklbl =
  let x = ref 0 in
  (fun () -> incr x; ("lbl__" ^(string_of_int !x)))

let rec change_exit_point lbl = function
  | [s,l,No_branch] -> [s,l,Goto lbl]
  | [ ( _,_,Label_expr (_,_) ) as e] -> e :: [mklbl (),[],Goto lbl]
  | (s,l,End)::tl -> (s,l,Goto lbl) :: (change_exit_point lbl tl)
  | hd::tl -> hd::(change_exit_point lbl tl)
  | [] -> []

let concat a b = match b with
  | [] -> a
  | (lbl,_,_) :: _ -> (change_exit_point lbl a) @ b
  

let initcode ic c =
  match ic,c with
  | [],_ -> c
  | _,[] -> assert false
  | (lblic,_,_) :: _, (lblc,_,_) :: _->
      begin
        let initialized = 250 in
        [
          "initcode",[],Goto_manual initialized;
          "initcode'", [], Goto lblic;
          "initcode_leave", [], Label_expr (initialized, lblc);
          "initcode_finish",[], Goto lblc;
        ] @ (change_exit_point "initcode_leave" ic) @ c
      end

let connardize_zerodiv c =
  let x = 254 in
  let y = 255 in
  concat
    [
      "connard_entry",
      [
        Int2;
        Set_dir ( Const 3 );
      ],
      End;
    ]
    (
      concat c
        [
          "connard_halt",
          [
            Var x =^ Const 0;
            Var y =^ Const 255;
            Var x =^ Var y /^ Var x;
            Set_dir ( Const 2 ) 
          ],
          End;
        ]
    )

let connardize_goto255 c =
  let y = 255 in
  concat
    [
      "connard_entry",
      [
        Int2;
        Set_dir ( Const 3 );
      ],
      End;
    ]
    (
      concat c
        [
          "connard_halt",
          [ Var y =^ Const 255; ],
          Goto_manual y;
        ]
    )

let connardize c =
  let x = 254 in
  let y = 255 in
  let z = 253 in
  let z' = 252 in
  let connard : code =
    [
      "connard_entry",
      [
        Int2;
        Set_dir ( Const 3 );
      ],
      Goto "start";
      "halt",
      [
        Var z' =^ Var z;
        Var z =^ Const 1 -^ Var z';
      ],
      Cond2 (Eq, "connard_zerodiv", "connard_manual255", Var z, Const 0);
      "connard_zerodiv",
      [
        Var x =^ Const 0;
        Var y =^ Const 255;
        Var x =^ Var y /^ Var x;
      ], Goto "connard_tail";
      "connard_manual255",
      [ Var y =^ Const 255; ],
      Goto_manual y;
      "connard_tail",
      [ Set_dir ( Const 2 ) ],
      End;
    ] @ c in
  let code = Array.of_list (build_blocks connard) in
  let o = open_out "/tmp/connard.gcc" in
  print (Format.formatter_of_out_channel o)
    {code; data = [||]; pc = 0; registers = [||];};
  close_out o

let () = connardize [ "start",[],Goto "halt" ]


let print_code f c =
  let code = Array.of_list (build_blocks c) in
  let o = open_out f in
  print (Format.formatter_of_out_channel o)
    {code; data = [||]; pc = 0; registers = [||];};
  close_out o

let () =
  print_code "/tmp/zerodiv.gcc" (connardize_zerodiv [ "t_start",[],End ] );
  print_code "/tmp/goto255.gcc"
    (connardize_goto255 ["t_start",[Var 18 =^ Const 0],End ] )
  
