open World

type tick = int

type ghost_tick_queue = (tick * int) list
type 'a lambda_tick_queue = (tick * ('a GccRun.runner * 'a GccRun.closure * 'a GccTypes.value)) list

type 'a running_world =
    { lambda_code : 'a GccRun.runner;
      mutable tick : tick;
      mutable ghost_tick_queue : ghost_tick_queue;
      mutable lambda_tick_queue : 'a lambda_tick_queue;
      mutable world : world;
      mutable ghosts_states : Ghc.ghc array;
      mutable pill_count : int;
      mutable number_of_ghost_eaten : int;
      mutable some_move : bool }

let run_lambda_code program closure (arg1,arg2)
    : ( 'a GccTypes.value * 'a GccTypes.value ) option =
  let _ = GccRun.get_count () in
  let res = GccRun.run_instrs' program closure [|arg1; arg2|] in
  let n = GccRun.get_count () in
  (* Printf.printf "count :%i\n%!" n; *)
  match res with
  | Some (GccTypes.TAG_CONS (v1, v2)) ->
     Some (v1,v2)
  | _ ->
     None

let print_dir ppf = function
  | Up -> Format.fprintf ppf "Up"
  | Down -> Format.fprintf ppf "Down"
  | Left -> Format.fprintf ppf "Left"
  | Right -> Format.fprintf ppf "Right"

let run_lambda (type t) (program: t GccRun.runner) closure (arg1,arg2)
    : ( direction option * t GccTypes.value ) =
  let res = run_lambda_code program closure (arg1,arg2) in
  match res with
  | Some (st, TAG_INT i) ->
    begin
      match direction_of_int i with
      | None ->
         Format.eprintf "Bad result@.";
         None, arg1
      | Some dir ->
         (* Format.eprintf "DIR %a@." print_dir dir; *)
         Some dir, st
    end
  | _ ->
     Format.eprintf "Bad result@.";
     (* Que faut il faire quand ça crash ? *)
     None, arg1

let next_position world dir =
  let pos = world.lambda_man.location in
  let dest =
    match dir with
    | Up -> { pos with y = pos.y - 1 }
    | Down -> { pos with y = pos.y + 1 }
    | Left -> { pos with x = pos.x - 1 }
    | Right -> { pos with x = pos.x + 1 }
  in
  let height = Array.length world.map in
  let width = Array.length world.map.(0) in
  let ok = dest.x >= 0 && dest.y >= 0
           && dest.x < width && dest.y < height in
  if ok
  then dest
  else pos

let move_lambda_man world dir =
  let orig_pos = world.lambda_man.location in
  let pos = next_position world dir in
  (* Format.eprintf "move %i, %i -> %i, %i@." orig_pos.x orig_pos.y pos.x pos.y; *)
  let new_pos, eating =
    match world.map.(pos.y).(pos.x) with
    | Lambda_man
    | Ghost
    | Empty -> pos, false
    | Wall -> orig_pos, false
    | Pill
    | Power_pill
    | Fruit -> pos, true
  in
  (* Format.eprintf "new_pos: %i, %i\teating: %b@." new_pos.x new_pos.y eating; *)
  let next_tick =
    if eating
    then Timecst.lambda_man_eating_walk_time
    else Timecst.lambda_man_walk_time in
  { world with
    lambda_man =
      { world.lambda_man with location = new_pos } },
  next_tick

let run_lambdas (type t) (state:t running_world) =
  let rec aux : t lambda_tick_queue -> t lambda_tick_queue = function
    | [] -> []
    | ((tick, (id, closure, value))) :: t ->
       if tick = state.tick
       then
         let t = aux t in
         state.some_move <- true;
         let arg = World.value_of_world state.world in
         let (dir, value) = run_lambda id closure (value, arg) in
         let dir =
           match dir with
           | None -> state.world.lambda_man.direction
           | Some dir -> dir
         in
         let world, next_time = move_lambda_man state.world dir in
         let next_time = next_time + state.tick in
         state.world <- world;
         (next_time, (id, closure, value)) :: t
       else
         let t = aux t in
         (tick, (id, closure, value)) :: t
  in
  let lambda_tick_queue = aux state.lambda_tick_queue in
  state.lambda_tick_queue <- lambda_tick_queue;
  state

let reachable = function
  | Wall -> false
  | Pill
  | Power_pill
  | Fruit
  | Lambda_man
  | Ghost
  | Empty -> true

let cell world pos =
  try
    world.map.(pos.y).(pos.x)
  with _ -> Wall

let set_cell world pos v =
  try
    world.map.(pos.y).(pos.x) <- v
  with _ -> ()

let world_cell world x y =
  try
    reachable world.map.(y).(x)
  with _ -> false

let opposite x y =
  match x, y with
  | Up, Down
  | Down, Up
  | Left, Right
  | Right, Left -> true
  | _ -> false

let move dir pos =
  match dir with
  | Up -> { pos with y = pos.y - 1 }
  | Down -> { pos with y = pos.y + 1 }
  | Left -> { pos with x = pos.x - 1 }
  | Right -> { pos with x = pos.x + 1 }

let change_ghost_dir world id pos dir =
  let orig_dir = world.ghosts.(id).direction in
  let up = world_cell world pos.x (pos.y - 1) in
  let down = world_cell world pos.x (pos.y + 1) in
  let left = world_cell world (pos.x - 1) pos.y in
  let right = world_cell world (pos.x + 1) pos.y in
  let check_180 dir =
    let dead_end = function
      | Up ->    not (down || left || right)
      | Down ->  not (up || left || right)
      | Left ->  not (up || down || right)
      | Right -> not (up || down || left)
    in
    not (opposite dir orig_dir) || dead_end dir
  in
  let up = up && check_180 Up in
  let down = down && check_180 Down in
  let left = left && check_180 Left in
  let right = right && check_180 Right in
  let legal_move = function
    | Up -> up
    | Down -> down
    | Left -> left
    | Right -> right
  in
  if legal_move dir then dir
  else if legal_move orig_dir then orig_dir
  else if up then Up
  else if right then Right
  else if down then Down
  else Left

let run_ghosts state =
  let rec aux state = function
    | [] -> [], state
    | (tick, id) :: t ->
       let code_id = id mod (Array.length state.ghosts_states) in
       let t, state = aux state t in
       let world = state.world in
       let pos = state.world.ghosts.(id).current_location in
       if tick = state.tick
       then
         let dir = Ghc.interp state.ghosts_states.(code_id) state.world id in
         Format.eprintf "run ghost: %i DIR %a@." id print_dir dir;
         state.some_move <- true;
         let dir = change_ghost_dir state.world id pos dir in
         let pos = move dir pos in
         world.ghosts.(id) <-
           { world.ghosts.(id)
           with
             direction = dir;
             current_location = pos };
         let dtick =
           let open Timecst in
           if world.lambda_man.vitality > 0
           then
             match code_id with
             | 0 -> fright_ghost_walk_time_0
             | 1 -> fright_ghost_walk_time_1
             | 2 -> fright_ghost_walk_time_2
             | _ -> fright_ghost_walk_time_3
           else
             match code_id with
             | 0 -> normal_ghost_walk_time_0
             | 1 -> normal_ghost_walk_time_1
             | 2 -> normal_ghost_walk_time_2
             | _ -> normal_ghost_walk_time_3
         in
         (tick + dtick, id) :: t, state
       else (tick, id) :: t, state
  in
  let ghost_tick_queue, state = aux state state.ghost_tick_queue in
  { state with ghost_tick_queue }

let level world =
  let width = Array.length world.map in
  let height = Array.length world.map.(0) in
  ((width * height) / 100) + 1

let fruit_score tick world =
  match level world with
  | 1 -> 100
  | 2 -> 200
  | 3 -> 500
  | 4 -> 500
  | 5 -> 700
  | 6 -> 700
  | 7 -> 1000
  | 8 -> 1000
  | 9 -> 2000
  | 10 -> 2000
  | 11 -> 3000
  | 12 -> 3000
  | _ -> 5000

exception Lost of int

let loose state =
  let world = state.world in
  let lambda_man =
    let lambda_man = world.lambda_man in
    if lambda_man.remaining_lives = 1
    then raise (Lost lambda_man.score)
    else { lambda_man with
           remaining_lives = lambda_man.remaining_lives - 1;
           location = lambda_man.starting_location;
           direction = Down }
  in
  let aux_ghost g =
    { g with
      current_location = g.starting_location;
      direction = Down }
  in
  let ghosts = Array.map aux_ghost world.ghosts in
  { state with world = { world with lambda_man; ghosts } }

let eat_ghost_score state i =
  let ghost = state.world.ghosts.(i) in
  if state.world.lambda_man.location = ghost.current_location
  then
    let score =
      match state.number_of_ghost_eaten with
      | 0 -> 200
      | 1 -> 400
      | 2 -> 800
      | _ -> 1600
    in
    let ghost =
      { ghost with
        vitality = Invisible;
        current_location = ghost.starting_location;
        direction = Down } in
    state.world.ghosts.(i) <- ghost;
    { state with
      world =
        { state.world with
          lambda_man =
            {state.world.lambda_man with
              score = state.world.lambda_man.score + score } };
      number_of_ghost_eaten = state.number_of_ghost_eaten + 1 }
  else state

type 'a result =
  | End of int
  | Continue of 'a running_world * bool

let run_step (type t) (state:t running_world) =
  let state = run_ghosts state in
  (* Format.eprintf "before: %i, %i@." *)
  (*                state.world.lambda_man.location.x *)
  (*                state.world.lambda_man.location.y; *)
  let state = run_lambdas state in
  (* Format.eprintf "after: %i, %i@." *)
  (*                state.world.lambda_man.location.x *)
  (*                state.world.lambda_man.location.y; *)
  let world = state.world in
  let world =
    { world with
      lambda_man =
        { world.lambda_man with
          vitality = max 0 (state.world.lambda_man.vitality - 1) } } in
  let world =
    let open Timecst in
    if state.tick = fruit1_appear
    then { world with fruit = Some (fruit1_expire - fruit1_appear) }
    else
      if state.tick = fruit2_appear
      then { world with fruit = Some (fruit2_expire - fruit2_appear) }
      else
        match world.fruit with
        | None -> world
        | Some time ->
           let time = time - 1 in
           if time = 0
           then { world with fruit = None }
           else { world with fruit = Some time }
  in
  let state = { state with world } in
  let state =
    let world = state.world in
    let { x; y } = world.lambda_man.location in
    let cell = cell world world.lambda_man.location in
    match cell with
    | Empty
    | Lambda_man
    | Ghost ->
       state
    | Pill ->
       let score = world.lambda_man.score + Timecst.pill_score in
       let lambda_man = { world.lambda_man with score } in
       world.map.(y).(x) <- Empty;
       let world = { world with lambda_man } in
       { state with world; pill_count = state.pill_count - 1 }
    | Power_pill ->
       let score = world.lambda_man.score + Timecst.power_pill_score in
       let vitality = Timecst.fright_mode_duration in
       let lambda_man = { world.lambda_man with vitality; score } in
       world.map.(y).(x) <- Empty;
       let world = { world with lambda_man } in
       { state with world; number_of_ghost_eaten = 0 }
    | Fruit ->
       begin
         match world.fruit with
         | None -> state
         | Some _ ->
            let lambda_man =
              { world.lambda_man with
                score = world.lambda_man.score + (fruit_score state.tick world) } in
            let world =
              { world with
                fruit = None;
                lambda_man } in
            { state with world }
       end
    | Wall -> assert false
  in
  let encounter =
    List.exists
      (fun ({ current_location }:ghost) ->
       current_location = state.world.lambda_man.location)
      (Array.to_list state.world.ghosts)
  in
  let fright_mode = state.world.lambda_man.vitality > 0 in
  let state =
    if encounter
    then
      if fright_mode
      then
        begin
          let ghosts = state.world.ghosts in
          let state = ref state in
          Array.iteri
            (fun i _ -> state := eat_ghost_score !state i)
            ghosts;
          !state
        end
      else loose state
    else state
  in
  (* Format.eprintf "pill_count: %i@." state.pill_count; *)
  if state.pill_count = 0
  then End state.world.lambda_man.score
  else
    let some_move = state.some_move in
    state.tick <- state.tick + 1;
    state.some_move <- false;
    Continue (state, some_move)

let init_lambda_man (type t) (code:t GccRun.runner) world =
  let state = GccRun.init_state' code in
  let value = World.value_of_world world in
  match run_lambda_code code (GccRun.entry_point code, state.env) (value, GccTypes.TAG_INT 133) with
  | Some (value, TAG_CLOSURE (point, frame)) ->
     (point, frame), (value:t GccTypes.value)
  | _ -> failwith "init_lambda_man failed"

(* let compare_ghost g1 g2 = *)
(*   let { x = x1; y = y1 } = g1.starting_location in *)
(*   let { x = x2; y = y2 } = g2.starting_location in *)
(*   if y1 < y2 *)
(*   then -1 *)
(*   else if y1 > y2 *)
(*   then 1 *)
(*   else if x1 < y2 *)
(*   then -1 *)
(*   else if x1 > y2 *)
(*   then 1 *)
(*   else 0 *)

let init_state ppf ~map ~ghost_files ~lambda_file =
  let lambda_code = GccParse.from_file lambda_file in
  (* let lambda_code = GccPrint.assemble lambda_code in *)
  (* let lambda_code = GccRun.int_runner lambda_code in *)
  let debug r v s c l =
    let f ppf =
      (* Format.fprintf ppf "debug: "; *)
      GccRun.simple_debug ppf r v s c l;
      Format.fprintf ppf "@."
    in
    f ppf;
    f Format.std_formatter
  in
  let trace =
    let b =
      try let _ = Sys.getenv "TRACE" in true
      with _ -> false
    in
    if b then Some (GccRun.simple_trace Format.err_formatter) else None
  in
  let lambda_code = GccRun.arg_runner ~debug ?trace lambda_code in
  let world = World.from_file map in
  let lambda_closure, lambda_man_state = init_lambda_man lambda_code world in
  let gh =
    Array.mapi (fun i _ -> (0, i)) world.ghosts
    |> Array.to_list in
  let ghosts_states = Array.map Ghc.from_file ghost_files in
  Format.printf "ghosts: %i@." (Array.length ghosts_states);
  { lambda_code;
    tick = 0;
    world;
    pill_count = World.pill_count world;
    number_of_ghost_eaten = 0;
    lambda_tick_queue = [0, (lambda_code, lambda_closure, lambda_man_state)];
    ghost_tick_queue = gh;
    ghosts_states;
    some_move = true }
